/*
 * OpenBase Connectivity for Ruby
 *
 * Copyright (C) 2004- SUZUKI Tetsuya <suzuki@spice-of-life.net>
 * It is a free software distributed under the BSD license.
 * 
 * $Id: openbase.m,v 1.39 2005/11/10 07:49:47 john Exp $
 */

#import <OpenBaseAPI/OpenBase.h>
#import <OpenBaseAPI/DBEntity.h>
#import <OpenBaseAPI/capi.h>
#import <ruby.h>


// 
// Data Type Definitions
//

#define OB_CHAR_LENGTH      4096
#define OB_MONEY_LENGTH     32
#define OB_DATE_LENGTH      32
#define OB_TIME_LENGTH      32
#define OB_DATETIME_LENGTH  32
#define OB_TIMESTAMP_LENGTH 32
#define OB_BLOB_KEY_LENGTH  8
#define OB_BOOLEAN_LENGTH   2

typedef struct _openbase {
  OpenBase *conn;
} openbase;

typedef struct _obresult {
  OpenBase *conn;
  VALUE infos;
  VALUE row;
  int count;  // column count
  int *types; // result column types
  void **buf; // buffer to bind result columns
  BOOL freed;
} obresult;

typedef struct _obcolumninfo {
  VALUE name;
  VALUE type;
  VALUE table;
} obcolumninfo;

static VALUE rb_cOpenBase;
static VALUE rb_cOpenBaseResult;
static VALUE rb_cOpenBaseColumnInfo;
static VALUE rb_cOpenBaseError;

#define GET_OPENBASE(obj,var) Data_Get_Struct(obj, openbase, var)
#define GET_RESULT(obj,var)   Data_Get_Struct(obj, obresult, var)
#define GET_INFO(obj,var)     Data_Get_Struct(obj, obcolumninfo, var)

#define GET_OPENBASE_WITH_CHECK(obj,var) \
    Data_Get_Struct(obj, openbase, var); \
    if ([var->conn isConnected] == NO) { \
        rb_raise(rb_cOpenBaseError, "The database is not connected."); \
    }

#define GET_RESULT_WITH_CHECK(obj,var) \
    Data_Get_Struct(obj, obresult, var); \
    if ([var->conn isConnected] == NO) { \
        rb_raise(rb_cOpenBaseError, "The database is not connected."); \
    }

#define INITPOOL    NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
#define RELEASEPOOL [pool release];

#define POOL(statement) \
    NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init]; \
    statement; [pool release];


// 
// Function Prototypes
//

static VALUE openbase_connect _((int argc, VALUE *argv, VALUE self));
static VALUE result_new       _((VALUE self, VALUE obconn));
static VALUE columninfo_new   _((VALUE self, VALUE name, VALUE type, VALUE table));


// 
// Methods
//

static void openbase_free( openbase *p )
{
  INITPOOL;
  if ([p->conn isConnected]) [p->conn invalidate];
  [p->conn release];
  RELEASEPOOL;

  free(p);
}


static void openbase_raise_error( int i )
{
  char msg[40];
  sprintf(msg, "unknown result column type - %d", i);
  rb_raise(rb_cOpenBaseError, msg);
}


static VALUE openbase_alloc( VALUE klass )
{
  openbase *ob = ALLOC(openbase);
  return Data_Wrap_Struct(klass, 0, openbase_free, ob);
}


static VALUE openbase_initialize( int argc, VALUE *argv, VALUE self )
{
  VALUE arg[4];
  openbase *ob;
  VALUE database, host, login, password;

  GET_OPENBASE(self, ob);

  POOL(ob->conn = [[OpenBase alloc] init]);
  rb_scan_args(argc, argv, "04", &arg[0], &arg[1], &arg[2], &arg[3]);

  if (!NIL_P(arg[0])) {
    Check_Type(arg[0], T_STRING);
    database = arg[0];
  } else {
    database = rb_str_new2("");
  }

  if (!NIL_P(arg[1])) {
    Check_Type(arg[1], T_STRING);
    host = arg[1];
  } else {
    host = rb_str_new2("localhost");
  }

  if (!NIL_P(arg[2])) {
    Check_Type(arg[2], T_STRING);
    login = arg[2];
  } else {
    login = rb_str_new2("admin");
  }

  if (!NIL_P(arg[3])) {
    Check_Type(arg[3], T_STRING);
    password = arg[3];
  } else {
    password = rb_str_new2("");
  }

  rb_funcall(self, rb_intern("connect"), 4, database, host, login, password);

  return Qnil;
}


static VALUE openbase_connect( int argc, VALUE *argv, VALUE self )
{
  VALUE arg[4];
  openbase *ob;
  char *database, *host, *login, *password;
  const char *msg;
  int return_code;

  GET_OPENBASE(self, ob);
  rb_scan_args(argc, argv, "04", &arg[0], &arg[1], &arg[2], &arg[3]);

  if (!NIL_P(arg[0])) {
    Check_Type(arg[0], T_STRING);
    database = STR2CSTR(arg[0]);
  } else {
    database = "";
  }

  if (!NIL_P(arg[1])) {
    Check_Type(arg[1], T_STRING);
    host = STR2CSTR(arg[1]);
  } else {
    host = "localhost";
  }

  if (!NIL_P(arg[2])) {
    Check_Type(arg[2], T_STRING);
    login = STR2CSTR(arg[2]);
  } else {
    login = "admin";
  }

  if (!NIL_P(arg[3])) {
    Check_Type(arg[3], T_STRING);
    password = STR2CSTR(arg[3]);
  } else {
    password = "";
  }

  INITPOOL;
  if ([ob->conn isConnected]) [ob->conn invalidate];
  if ([ob->conn connectToDatabase:database onHost:host login:login
         password:password return:&return_code] == NO) {
    msg = [ob->conn connectErrorMessage:return_code];
    RELEASEPOOL;

    rb_raise(rb_cOpenBaseError, msg);
  }
  RELEASEPOOL;

  return Qnil;
}


static VALUE openbase_invalidate( VALUE self )
{
  openbase *ob;

  GET_OPENBASE(self, ob);
  INITPOOL;
  if ([ob->conn isConnected]) [ob->conn invalidate];
  RELEASEPOOL;

  return Qnil;
}


static VALUE openbase_is_connected( VALUE self )
{
  openbase *ob;
  BOOL flag;

  GET_OPENBASE(self, ob);
  POOL(flag = [ob->conn isConnected]);
  if (flag) {
    return Qtrue;
  } else {
    return Qfalse;
  }
}


static VALUE openbase_encoding( VALUE self )
{
  openbase *ob;
  VALUE str;
  GET_OPENBASE_WITH_CHECK(self, ob);

  INITPOOL;
  str = rb_str_new2([ob->conn databaseEncodingName]);
  RELEASEPOOL;

  return str;
}


static VALUE openbase_tables( VALUE self )
{
  openbase *ob;
  VALUE ary;
  id e, name;
  const char *cname;

  GET_OPENBASE_WITH_CHECK(self, ob);
  INITPOOL;
  ary = rb_ary_new();
  e = [[DBEntity tableNamesWithConnection:ob->conn] objectEnumerator];
  while (name = [e nextObject]) {
    cname = [name cStringUsingEncoding:[ob->conn databaseEncoding]];
    rb_ary_push(ary, rb_str_new2(cname));
  }
  RELEASEPOOL;
  return ary;
}


static VALUE openbase_columns( VALUE self, VALUE table )
{
  openbase *ob;
  VALUE ary;
  id entity, e, name, tname;
  const char *cname;

  GET_OPENBASE_WITH_CHECK(self, ob);
  INITPOOL;
  ary = rb_ary_new();
  tname = [NSString stringWithCString:RSTRING(table)->ptr
                    encoding:[ob->conn databaseEncoding]];
  entity = [DBEntity entityForTable:tname connection:ob->conn];
  e = [[entity columns] objectEnumerator];
  while (name = [e nextObject]) {
    cname = [name cStringUsingEncoding:[ob->conn databaseEncoding]];
    rb_ary_push(ary, rb_str_new2(cname));
  }
  RELEASEPOOL;
  return ary;
}


static VALUE openbase_relationships( VALUE self, VALUE table )
{
  openbase *ob;
  VALUE ary;
  id e, name, tname;
  const char *cname;

  GET_OPENBASE_WITH_CHECK(self, ob);
  INITPOOL;
  ary = rb_ary_new();
  tname = [NSString stringWithCString:RSTRING(table)->ptr
                    encoding:[ob->conn databaseEncoding]];
  e = [[DBEntity relationshipsForTable:tname connection:ob->conn] objectEnumerator];
  while (name = [e nextObject]) {
    cname = [name cStringUsingEncoding:[ob->conn databaseEncoding]];
    rb_ary_push(ary, rb_str_new2(cname));
  }
  RELEASEPOOL;
  return ary;
}


static VALUE openbase_database( VALUE self )
{
  openbase *ob;
  VALUE str;

  GET_OPENBASE_WITH_CHECK(self, ob);
  INITPOOL;
  str = rb_str_new2([ob->conn databaseName]);
  RELEASEPOOL;

  return str;
}


static VALUE openbase_host( VALUE self )
{
  openbase *ob;
  VALUE str;

  GET_OPENBASE_WITH_CHECK(self, ob);
  INITPOOL;
  str = rb_str_new2([ob->conn hostName]);
  RELEASEPOOL;

  return str;
}


static VALUE openbase_login( VALUE self )
{
  openbase *ob;
  VALUE str;

  GET_OPENBASE_WITH_CHECK(self, ob);
  INITPOOL;
  str = rb_str_new2([ob->conn loginName]);
  RELEASEPOOL;

  return str;
}


static VALUE openbase_password( VALUE self )
{
  openbase *ob;
  VALUE str;

  GET_OPENBASE_WITH_CHECK(self, ob);
  INITPOOL;
  str =  rb_str_new2([ob->conn password]);
  RELEASEPOOL;

  return str;
}


static VALUE openbase_begin( VALUE self )
{
  openbase *ob;
  const char *msg = NULL;

  GET_OPENBASE_WITH_CHECK(self, ob);
  INITPOOL;
  if ([ob->conn beginTransaction] == NO) msg = [ob->conn serverMessage];
  RELEASEPOOL;

  if (msg != NULL) rb_raise(rb_cOpenBaseError, msg);

  return Qnil;
}


static VALUE openbase_commit( VALUE self )
{
  openbase *ob;
  const char *msg = NULL;

  GET_OPENBASE_WITH_CHECK(self, ob);
  INITPOOL;
  if ([ob->conn commitTransaction] == NO) msg = [ob->conn serverMessage];
  RELEASEPOOL;

  if (msg != NULL) rb_raise(rb_cOpenBaseError, msg);

  return Qnil;
}


static VALUE openbase_rollback( VALUE self )
{
  openbase *ob;
  const char *msg = NULL;

  GET_OPENBASE_WITH_CHECK(self, ob);
  INITPOOL;
  if ([ob->conn rollbackTransaction] == NO) msg = [ob->conn serverMessage];
  RELEASEPOOL;

  if (msg != NULL) rb_raise(rb_cOpenBaseError, msg);

  return Qnil;
}


static VALUE openbase_in_progress( VALUE self )
{
  openbase *ob;
  BOOL flag;

  GET_OPENBASE_WITH_CHECK(self, ob);
  POOL(flag = [ob->conn transactionInProgress]);

  if (flag == YES) {
    return Qtrue;
  } else {
    return Qfalse;
  }
}


static VALUE openbase_transaction( VALUE self )
{
  openbase *ob;
  int state;

  GET_OPENBASE_WITH_CHECK(self, ob);

  INITPOOL;

  if ([ob->conn transactionInProgress])
    [ob->conn commitTransaction];

  [ob->conn beginTransaction];

  rb_protect(rb_yield, self, &state);
  if (state != 0) {
    [ob->conn rollbackTransaction];
    RELEASEPOOL;
    rb_exc_raise(ruby_errinfo);
  }

  [ob->conn commitTransaction];

  RELEASEPOOL;

  return Qnil;
}


static VALUE openbase_unique_row_id( int argc, VALUE *argv, VALUE self )
{
  openbase *ob;
  VALUE arg[2];
  char *table, *column, *row_id;

  GET_OPENBASE_WITH_CHECK(self, ob);
  rb_scan_args(argc, argv, "02", &arg[0], &arg[1]);
  
  if (!NIL_P(arg[0])) {
    Check_Type(arg[0], T_STRING);
    table = STR2CSTR(arg[0]);
  }

  INITPOOL;
  if (!NIL_P(arg[1])) {
    Check_Type(arg[1], T_STRING);
    column = STR2CSTR(arg[1]);
    row_id = (char *)[ob->conn uniqueRowIdForTable:table column:column];
  } else {
    row_id = (char *)[ob->conn uniqueRowIdForTable:table];
  }
  RELEASEPOOL;

  return rb_str_new2(row_id);
}


static VALUE openbase_execute( VALUE self, VALUE statement )
{
  openbase *ob;
  VALUE *result;
  const char *msg = NULL;

  Check_Type(statement, T_STRING);
  GET_OPENBASE_WITH_CHECK(self, ob);

  INITPOOL;
  [ob->conn makeCommand:RSTRING(statement)->ptr];
  if ([ob->conn executeCommand] == NO) msg = [ob->conn serverMessage];

  if (msg != NULL) rb_raise(rb_cOpenBaseError, msg);

  RELEASEPOOL;

  return rb_funcall(rb_cOpenBaseResult, rb_intern("new"), 1, self);
}


static void result_mark( obresult *p )
{
  rb_gc_mark(p->infos);
  rb_gc_mark(p->row);
}


static void result_free( obresult *p )
{
  if (!p->freed) {
    free(p->types);
    free(p->buf);
    POOL([p->conn release]);
  }
  free(p);
}


static VALUE result_new( VALUE self, VALUE obconn )
{
  obresult *result;
  OpenBase *ob;
  int i, count, num;
  char *type, *name, *table, msg[40];
  VALUE info, value;
  void **buf;

  ob = ((openbase *)(RDATA(obconn)->data))->conn;
  result = ALLOC(obresult);
  result->freed = NO;
  result->conn  = [ob retain];
  result->infos = rb_ary_new();
  result->row   = rb_ary_new();

  INITPOOL;
  count = [ob resultColumnCount];
  result->count = count;
  result->types = malloc(sizeof(int) * count);
  buf = malloc(sizeof(void *) * count);
  result->buf = buf;
  for (i = 0; i < count; i++) {
    num = [ob resultColumnType:i];
    result->types[i] = num;
    switch (num) {
    case OBTYPE_CHAR:
      type = "char";
      buf[i] = malloc(sizeof(char) * OB_CHAR_LENGTH);
      [ob bindString:(char *)buf[i]];
      break;
    case OBTYPE_INT:
      type = "int";
      buf[i] = malloc(sizeof(int));
      [ob bindInt:(int *)buf[i]];
      break;
    case OBTYPE_FLOAT:
      type = "float";
      buf[i] = malloc(sizeof(double));
      [ob bindDouble:(double *)buf[i]];
      break;
    case OBTYPE_LONG:
      type = "long";
      buf[i] = malloc(sizeof(long));
      [ob bindLong:(long *)buf[i]];
      break;
    case OBTYPE_MONEY:
      type = "money";
      buf[i] = malloc(sizeof(char) * OB_MONEY_LENGTH);
      [ob bindString:(char *)buf[i]];
      break;
    case OBTYPE_DATE:
      type = "date";
      buf[i] = malloc(sizeof(char) * OB_DATE_LENGTH);
      [ob bindString:(char *)buf[i]];
      break;
    case OBTYPE_TIME:
      type = "time";
      buf[i] = malloc(sizeof(char) * OB_TIME_LENGTH);
      [ob bindString:(char *)buf[i]];
      break;
    case OBTYPE_OBJECT:
      type = "object";
      buf[i] = malloc(sizeof(char) * OB_BLOB_KEY_LENGTH);
      [ob bindString:(char *)buf[i]];
      break;
    case OBTYPE_DATETIME:
      type = "datetime";
      buf[i] = malloc(sizeof(char) * OB_DATETIME_LENGTH);
      [ob bindString:(char *)buf[i]];
      break;
    case OBTYPE_LONGLONG:
      type = "longlong";
      buf[i] = malloc(sizeof(long long));
      [result->conn bindLongLong:(long long *)buf[i]];
      break;
    case OBTYPE_BOOLEAN:
      type = "boolean";
      buf[i] = malloc(sizeof(char) * OB_BOOLEAN_LENGTH);
      [ob bindString:(char *)buf[i]];
      break;
    case OBTYPE_BINARY:
      type = "binary";
      buf[i] = malloc(sizeof(char) * OB_BLOB_KEY_LENGTH);
      [ob bindString:(char *)buf[i]];
      break;
    case OBTYPE_OBJECT_TEXT:
      type = "text";
      buf[i] = malloc(sizeof(char) * OB_BLOB_KEY_LENGTH);
      [ob bindString:(char *)buf[i]];
      break;
    case OBTYPE_TIMESTAMP:
      type = "timestamp";
      buf[i] = malloc(sizeof(char) * OB_TIMESTAMP_LENGTH);
      [ob bindString:(char *)buf[i]];
      break;
    default:
      openbase_raise_error([ob resultColumnType:i]);
    }
    name = (char *)[ob resultColumnName:i];
    table = (char *)[ob resultTableName:i];
    info = columninfo_new(rb_cOpenBaseColumnInfo,
                          rb_str_new2(name),
                          rb_str_new2(type),
                          rb_str_new2(table));
    rb_ary_push(result->infos, info);
  }
  RELEASEPOOL;
  return Data_Wrap_Struct(self, result_mark, result_free, result);
}


static VALUE result_object_free( VALUE self )
{
  obresult *result;
  char msg[40];

  GET_RESULT(self, result);
  if (result->freed) {
    sprintf(msg, "%s is already freed.", RSTRING(rb_inspect(self))->ptr);
    rb_raise(rb_cOpenBaseError, msg);
  }
  free(result->types);
  free(result->buf);
  POOL([result->conn release]);
  result->freed = YES;
  return Qnil;
}


static VALUE result_fetch( VALUE self )
{
  obresult *result;
  int i;
  NSData *data;
  BOOL nextRow;
  char *boolS;
  VALUE boolVal;

  GET_RESULT_WITH_CHECK(self, result);
  INITPOOL;
  rb_ary_clear(result->row);
  if (nextRow = [result->conn nextRow]) {
    for (i = 0; i < result->count; i++) {
      switch (result->types[i]) {
      case OBTYPE_CHAR:
      case OBTYPE_DATE:
      case OBTYPE_TIME:
      case OBTYPE_DATETIME:
      case OBTYPE_TIMESTAMP:
      case OBTYPE_MONEY:
        rb_ary_push(result->row, rb_str_new2((char *)result->buf[i]));
        break;
      case OBTYPE_BOOLEAN:
        boolS = (char *)result->buf[i];
        if (strlen(boolS) == 0)
          boolVal = Qnil;
        else if (strcmp(boolS, "1") == 0)
          boolVal = Qtrue;
        else
          boolVal = Qfalse;
        rb_ary_push(result->row, boolVal);
        break;
      case OBTYPE_INT:
        rb_ary_push(result->row, INT2FIX(*(int *)result->buf[i]));
        break;
      case OBTYPE_LONG:
        rb_ary_push(result->row, LONG2NUM(*(long *)result->buf[i]));
        break;
      case OBTYPE_LONGLONG:
        rb_ary_push(result->row, LL2NUM(*(long long *)result->buf[i]));
        break;
      case OBTYPE_FLOAT:
        rb_ary_push(result->row, rb_float_new(*(double *)result->buf[i]));
        break;
      case OBTYPE_BINARY:
      case OBTYPE_OBJECT:
      case OBTYPE_OBJECT_TEXT:
        if (data = [result->conn retrieveBinary:(char *)result->buf[i]])
          rb_ary_push(result->row, rb_str_new([data bytes], [data length]));
        else
          rb_ary_push(result->row, rb_str_new2(""));
        break;
      default:
        openbase_raise_error([result->conn resultColumnType:i]);
      }
    }
  }
  RELEASEPOOL;
  if (nextRow)
    return result->row;
  else
    return Qnil;
}


static VALUE result_each( VALUE self )
{
  VALUE row;

  while ((row = result_fetch(self)) != Qnil) {
    rb_yield(row);
  }

  return self;
}


static VALUE result_column_infos( VALUE self )
{
  obresult *result;
  GET_RESULT(self, result);
  return result->infos;
}


static VALUE result_column_count( VALUE self )
{
  obresult *result;
  GET_RESULT(self, result);
  return INT2FIX(result->count);
}


static VALUE result_rows_affected( VALUE self )
{
  obresult *result;
  VALUE rows;

  GET_RESULT(self, result);
  INITPOOL;
  rows = INT2FIX([result->conn rowsAffected]);
  RELEASEPOOL;

  return rows;
}


static void columninfo_mark( obcolumninfo *p )
{
  rb_gc_mark(p->name);
  rb_gc_mark(p->type);
  rb_gc_mark(p->table);
}


static VALUE columninfo_new( VALUE self, VALUE name, VALUE type, VALUE table )
{
  obcolumninfo *info;

  info = ALLOC(obcolumninfo);
  Check_Type(name, T_STRING);
  Check_Type(type, T_STRING);
  Check_Type(table, T_STRING);
  info->name = name;
  info->type = type;
  info->table = table;

  return Data_Wrap_Struct(self, columninfo_mark, -1, info);
}


static VALUE columninfo_name( VALUE self )
{
  obcolumninfo *info;
  GET_INFO(self, info);
  return info->name;
}


static VALUE columninfo_type( VALUE self )
{
  obcolumninfo *info;
  GET_INFO(self, info);
  return info->type;
}


static VALUE columninfo_table( VALUE self )
{
  obcolumninfo *info;
  GET_INFO(self, info);
  return info->table;
}


//
// Initialization
//

void Init_openbase(void)
{
  VALUE rb_cOpenBase;

  rb_cOpenBase = rb_define_class("OpenBase", rb_cObject);
  rb_cOpenBaseResult = rb_define_class("OpenBaseResult", rb_cObject);
  rb_cOpenBaseColumnInfo = rb_define_class("OpenBaseColumnInfo", rb_cObject);
  rb_cOpenBaseError = rb_define_class("OpenBaseError", rb_eStandardError);

  // OpenBase
  rb_define_alloc_func(rb_cOpenBase, openbase_alloc);
  rb_define_private_method(rb_cOpenBase, "initialize", openbase_initialize, -1);

  rb_define_method(rb_cOpenBase, "database", openbase_database, 0);
  rb_define_method(rb_cOpenBase, "host", openbase_host, 0);
  rb_define_method(rb_cOpenBase, "login", openbase_login, 0);
  rb_define_method(rb_cOpenBase, "password", openbase_password, 0);
  rb_define_method(rb_cOpenBase, "encoding", openbase_encoding, 0);

  rb_define_method(rb_cOpenBase, "tables", openbase_tables, 0);
  rb_define_method(rb_cOpenBase, "columns", openbase_columns, 1);
  rb_define_method(rb_cOpenBase, "relationships", openbase_relationships, 1);

  rb_define_method(rb_cOpenBase, "connect", openbase_connect, -1);
  rb_define_method(rb_cOpenBase, "invalidate", openbase_invalidate, 0);
  rb_define_method(rb_cOpenBase, "connect?", openbase_is_connected, 0);

  rb_define_method(rb_cOpenBase, "begin", openbase_begin, 0);
  rb_define_method(rb_cOpenBase, "commit", openbase_commit, 0);
  rb_define_method(rb_cOpenBase, "rollback", openbase_rollback, 0);
  rb_define_method(rb_cOpenBase, "transaction?", openbase_in_progress, 0);
  rb_define_method(rb_cOpenBase, "transaction", openbase_transaction, 0);
  rb_define_method(rb_cOpenBase, "execute", openbase_execute, 1);

  rb_define_method(rb_cOpenBase, "unique_row_id", openbase_unique_row_id, -1);


  // OpenBaseResult
  rb_define_singleton_method(rb_cOpenBaseResult, "new", result_new, 1);
  rb_include_module(rb_cOpenBaseResult, rb_mEnumerable);

  rb_define_method(rb_cOpenBaseResult, "column_count", result_column_count, 0);
  rb_define_method(rb_cOpenBaseResult, "rows_affected", result_rows_affected, 0);
  rb_define_method(rb_cOpenBaseResult, "column_infos", result_column_infos, 0);
  rb_define_method(rb_cOpenBaseResult, "fetch", result_fetch, 0);
  rb_define_method(rb_cOpenBaseResult, "each", result_each, 0);
  rb_define_method(rb_cOpenBaseResult, "free", result_object_free, 0);


  // OpenBaseColumn
  rb_define_singleton_method(rb_cOpenBaseColumnInfo, "new", columninfo_new, 3);

  rb_define_method(rb_cOpenBaseColumnInfo, "name", columninfo_name, 0);
  rb_define_method(rb_cOpenBaseColumnInfo, "type", columninfo_type, 0);
  rb_define_method(rb_cOpenBaseColumnInfo, "table", columninfo_table, 0);
}

